// Copyright (c) 2000 National Semiconductor Corporation, All Rights Reserved.

/* This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
   for more details. 

   You should have received a copy of the GNU General Public License along with this program; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

/**********************************************************************
*h*
*
*   Name:               nsm.h
* 
*   Description:
*               HSM related data structures. 
*
*   Author:  NSC
*
*   History:  
*
*h*   
**********************************************************************/

#ifndef _NSMCONTEXT_H
#define _NSMCONTEXT_H

#define NsmGetTime()  jiffies

#define NsmZeroMemory(MacStats, Size) memset((PVOID) (MacStats), 0, (Size))

#ifdef IOMAPPED_IO

#define NsmRegRead8(pNsm, ioaddr, pRegVal)  *(pRegVal) = inb((unsigned)(ioaddr))

#define NsmRegRead16(pNsm, ioaddr, pRegVal)  *(pRegVal) = inw((unsigned)(ioaddr))

#define NsmRegRead32(pNsm, ioaddr, pRegVal)  *(pRegVal) = inl((unsigned)(ioaddr))

#define NsmRegWrite8(pNsm,ioaddr,data)  outb((data),(unsigned)(ioaddr))

#define NsmRegWrite16(pNsm,ioaddr,data)  outw((data), (unsigned)(ioaddr))

#define NsmRegWrite32(pNsm,ioaddr, data)  outl((data),(unsigned)(ioaddr))

#elif MEMMAPPED_IO

#define NsmRegRead8(pNsm, ioaddr, pRegVal)  *(pRegVal) = readb((unsigned)(ioaddr))

#define NsmRegRead16(pNsm, ioaddr, pRegVal)  *(pRegVal) = readw((unsigned)(ioaddr))

#define NsmRegRead32(pNsm, ioaddr, pRegVal)  *(pRegVal) = readl((unsigned)(ioaddr))

#define NsmRegWrite8(pNsm,ioaddr,data)  writeb((data),(unsigned)(ioaddr))

#define NsmRegWrite16(pNsm,ioaddr,data)  writew((data), (unsigned)(ioaddr))

#define NsmRegWrite32(pNsm,ioaddr, data)  writel((data),(unsigned)(ioaddr))

#endif


#define NsmCompare(SrcMac, DestMac, len) CMP_ETH_ADDR((SrcMac),(DestMac),(len))

#define NsmResetComplete(pAdapter)
#define NsmIndicateLinkStatus(pAdapter)


/* Nsmcontext structure */
/* This structure is not complete. Lots of things are to be added */
typedef struct NsmInfo
{
        struct _AdapterContext AdpCtxt;

        struct enet_statistics *pAdapterStats;
        struct device *next;
        struct MacAddr *pMultiList;
        struct device * dev;

} NsmContext;

typedef struct _EthHeader {

        UCHAR SrcAddr[6];
        UCHAR DstAddr[6];
        UINT size;

} EthHeader;

/* Tunable parameters */
#define BUF_SIZE_MAX  2048
#define TX_QUEUE_SZ 200
#define RX_QUEUE_SZ 100 

/* #define OK 0
#define ERROR -1 */
#define TRUE 1
#define FALSE 0
#define MAX_PRIORITY_QUEUES  1 
#define DUPED 10

#define DRV_NAME "dp83820"
#define DEV_STARTED 1
#define ETH_MAX_PKT_SIZE 2000
#define MIN_PKT_SIZE 64
#define MAC_ADDR_LEN 6
#define PCI_DEVICE_ID_83820 0x0022
#define BITNO 1

#define PCI_IO_SIZE 256 
#define PCI_MEM_SIZE 4096 

#define DEV_INTERRUPTED 1
#define DEV_SERVICED_INTERRUPT 0
#define SLEEP_STATE 2
#define HEAD_SPACE 2 

#define REPL_BM 100

#define NSMMAPPKTS
#define NSMGETNUMFRAGS

#define CACHE_LINE_SIZE 512

/* Linux Network driver interface routines */

extern int DrvProbe (struct device *dev);
STATIC int DrvOpen (struct device *dev);
STATIC int DrvClose (struct device *dev);
STATIC int DrvSend (struct sk_buff *skb, struct device *dev);
STATIC VOID DrvSetMulticastList (struct device *dev);
STATIC int DrvIoctl( struct device *dev, struct ifreq *rq, int cmd );
STATIC struct enet_statistics* DrvGetStatistics(struct device *dev);
STATIC VOID DrvIsr (int irq, VOID *dev_id, struct pt_regs *regs);

STATIC int ChangeMtu(struct device *dev, int NewMtu);



/* Driver private routines */

STATIC int init_module(VOID);
STATIC VOID cleanup_module(VOID);
STATIC VOID NsmTxCompleted(AdapterContext *pNsc, HsmTCB *TcbFirst, HsmTCB *TcbLast);

STATIC VOID NsmRxPackets (AdapterContext *pNsc, HsmRCB *RcbStart, HsmRCB *RcbEnd,
        UCHAR  Priority);
STATIC int ChangeMtu(struct device *dev, int NewMtu);

STATIC UCHAR  	NsmMallocContiguous(VOID *pNsm, 
	AddrStruct *pAddrStruct, USHORT Alignment);
STATIC VOID 	NsmFreeContiguous(VOID *pNsm, AddrStruct *pAddrStruct);
STATIC UCHAR  	NsmAllocatePacket(VOID *pNsm, HsmRCB *pRcb, 
					AddrStruct *pAddr);
STATIC VOID 	NsmDeAllocatePacket(VOID *pNsm, AddrStruct *pAddr, 
	HsmRCB *pRcb );

inline VOID *   NsmMalloc(UINT size);	
STATIC UCHAR IsBrdCastOrMultiCast( struct device *dev, UCHAR *DestMac);
VOID CreatePacketStruct(struct sk_buff *skb, PktStruct *pPkt);
VOID CreatePacketInfo(struct sk_buff *skb, PktInfo *pPkt);

inline UCHAR CMP_ETH_ADDR(UCHAR *SrcMac,UCHAR *DestMac, UINT Len);

#ifdef CLONE
#define NsmMarkRcbFree(pRcb)  \
{ \
	struct sk_buff *skb = ((struct sk_buff *)((pRcb)->pOsCtrlBlk));  \
	if((((skb_datarefp(skb))->counter) == 1) && ((pRcb)->Status == DUPED)) \
        (pRcb)->Status = RCB_FREE; \
}
#else
#define NsmMarkRcbFree(pRcb)  
#endif



/*
#define NsmMarkRcbFree(pRcb) \
if(((((struct sk_buff *)(pRcb->pOsCtrlBlk))->users) == 1) && \
	(pRcb->Status == DUPED)) \
        pRcb->Status = RCB_FREE
*/

#define NsmFree(ptr,size) kfree(ptr)

inline VOID *NsmCreateLock(); 

inline VOID NsmAcquireLock(AdapterContext *pAdapter,VOID *LockId); 

inline VOID NsmReleaseLock(AdapterContext *pNsm, VOID *LockId); 

#define NsmDestroyLock(pNsm, LockId) kfree(LockId)

#ifdef _DBG
#define NsmSleep(pNsm, SleepVal) mdelay(SleepVal);
#else
#define NsmSleep(pNsm, SleepVal) 
#endif

#define NsmCopy(src, dest, MAC_ADDR_LEN) \
	memcpy((dest),(src),MAC_ADDR_LEN)	

#define NsmDbgMsg(x) printk(KERN_INFO x)

#define NsmDbgMsg1(x,a1) printk(KERN_INFO x,a1)

#define NsmDbgMsg2(x,a1,a2) printk(KERN_INFO x,a1,a2)

#define NsmDbgMsg3(x,a1,a2,a3) printk(KERN_INFO x,a1,a2,a3)

#define NsmDbgMsg4(x,a1,a2,a3,a4) printk(KERN_INFO x,a1,a2,a3,a4)

inline VOID PrintContents(UCHAR *start, UINT len);

#endif _NSMCONTEXT_H

